package com.me.pacman.entity;

import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Vector2;
import com.me.pacman.entity.ai.Behaviour;
import com.me.pacman.entity.ai.ReturnToBase;
import com.me.pacman.entity.ai.Target;
import com.me.pacman.entity.path.EnterGhostHousePath;
import com.me.pacman.entity.path.ExitGhostHousePath;
import com.me.pacman.level.LevelTile;
import com.me.pacman.state.PlayState;

import java.util.ArrayList;

public class Ghost extends MovableEntity {

    public static final float GHOST_SPEED = 7.03f;
    public static final float GHOST_TUNNEL_SPEED = GHOST_SPEED / 2;
    public static final float EYES_SPEED = 15f;

    public static final Direction[] GHOST_ORDER = { Direction.UP, Direction.LEFT, Direction.DOWN, Direction.RIGHT };

    private TextureRegion[][] sprite;
    private int spriteIndex;
    private int counter = 0;

    private PlayState state;

    public Behaviour currentBehaviour;

    public Behaviour chaseBehaviour;
    public Behaviour scatterBehaviour;
    public Behaviour frightBehaviour;

    public boolean caught;

    public Ghost(PlayState state, float x, float y, Direction direction, int spriteIndex,
                 Behaviour chaseBehaviour, Behaviour scatterBehaviour) {
        super(state, x, y, GHOST_SPEED, true, direction, 0.1f);
        this.state = state;
        this.spriteIndex = spriteIndex;
        this.chaseBehaviour = chaseBehaviour;
        this.scatterBehaviour = scatterBehaviour;
        this.frightBehaviour = new FrightenedBehaviour(state);
        this.caught = false;
        sprite = state.getGame().assets.ghosts;
    }

    @Override
    public TextureRegion getSprite() {
        if (currentBehaviour instanceof ReturnToBase) {
            return null;
        } else if (state.frightTimer > 0 && !caught) {
            if (state.frightTimer > 2) {
                // Render the blue scared ghost
                return sprite[0][counter % 2];
            } else {
                // Render the flashing white ghost
                return sprite[0][(int) (state.frightTimer * 5) % 2 == 0? 2 : 0 + counter % 2];
            }
        } else {
            return sprite[2 + (counter % 2)][spriteIndex];
        }
    }

    @Override
    public void render(SpriteBatch batch, int offsetX, int offsetY) {
        super.render(batch, offsetX, offsetY);

        // draw eyes so the ghost can see
        if (state.frightTimer <= 0 || caught)
            batch.draw(sprite[1][currDirection.ordinal()], (int) (pos.x * 8) + (offsetX - 8), (pos.y * 8) + (offsetY - 8));
    }

    @Override
    public void update(float dt) {
        super.update(dt);

        if (age % 20 == 0) {
            counter++;
        }

        if (currentPath != null) {
            if (!currentPath.finished()) {
                return;
            }

            if (currentPath instanceof EnterGhostHousePath) {
                currentPath = new ExitGhostHousePath(pos);
                currentBehaviour = null;
                speed = GHOST_SPEED;
            } else if (currentPath instanceof ExitGhostHousePath) {
                currentBehaviour = chaseBehaviour;
                currentPath = null;
            }
        }

        if (currentBehaviour == null) {
            return;
        }

        Direction nextDirection = getNextDirection();

        if (!canMove) {
            // we're stuck somewhere, let's change directions to a new valid direction
            // check if we already have a nextDirection which will get us unstuck
            if (nextDirection != null) {
                Vector2 adjacent = nextDirection.getVector().add(pos);
                LevelTile adjTile = state.level.getTile(adjacent);
                if (adjTile != null && adjTile.isPassable()) {
                    return;
                }
            }
            // we didn't so check other directions
            for (Direction dir : GHOST_ORDER) {
                if (dir.isOpposite(currDirection) || dir == currDirection) {
                    // don't just turn around or keep going our current direction (since we'll continue being stuck)
                    continue;
                }
                Vector2 adjacent = dir.getVector().add(pos);
                LevelTile adjTile = state.level.getTile(adjacent);
                if (adjTile != null && adjTile.isPassable()) {
                    setNextDirection(dir);
                    return;
                }
            }
        }

        if (nextDirection != null) {
            return;
        }

        Target target = currentBehaviour.getTarget();
        if (target == null) {
            // no target, carry on current course
            return;
        }

        if (currentBehaviour instanceof ReturnToBase && target.targetReached(pos)) {
            currentPath = new EnterGhostHousePath(pos);
            return;
        }

        // Rudimentary path finding
        // Check all possible turns of the tile immediately in front of us and turn the direction that is closest
        // to our target.
        Vector2 ahead = new Vector2((int) pos.x, (int) pos.y).add(currDirection.getVector());
        float shortest = Float.MAX_VALUE;
        for (Direction dir : GHOST_ORDER) {
            if (dir.isOpposite(currDirection)) {
                continue;
            }
            Vector2 adjacent = dir.getVector().add(ahead);
            LevelTile nextTile = state.level.getTile(adjacent);
            if (nextTile != null && nextTile.isPassable()) {
                // compute distance to target
                float d = target.distance_sqr(adjacent);
                if (d < shortest) {
                    shortest = d;
                    nextDirection = dir;
                }
            }
        }
        setNextDirection(nextDirection);
    }

    private class FrightenedBehaviour extends Behaviour {

        private Target target;

        public FrightenedBehaviour(PlayState state) {
            super(state);
        }

        @Override
        public Target getTarget() {
            if (target != null && !target.targetReached(pos)) {
                return target;
            }

            ArrayList<Vector2> adjacentTiles = new ArrayList<>(3);
            ArrayList<Direction> possibleTurns = new ArrayList<>(3);
            Vector2 ahead = new Vector2((int) pos.x, (int) pos.y).add(currDirection.getVector());

            for (Direction dir : GHOST_ORDER) {
                if (dir.isOpposite(currDirection)) {
                    // Don't consider going backwards
                    continue;
                }
                Vector2 adjacent = dir.getVector().add(ahead);
                LevelTile nextTile = state.level.getTile(adjacent);
                if (nextTile != null && nextTile.isPassable()) {
                    adjacentTiles.add(adjacent);
                    possibleTurns.add(dir);
                }
            }

            if (possibleTurns.size() == 0) {
                // No possible turns means no valid tiles ahead, which would be the tunnel.
                target = null;
            } else if (possibleTurns.size() == 1) {
                if (possibleTurns.get(0) == currDirection) {
                    // can only go straight, so no target
                    target = null;
                } else {
                    // turn the only possible direction
                    target = new Target(adjacentTiles.get(0));
                }
            } else {
                // n-way intersection, pick random direction
                int rand = state.random.nextInt(adjacentTiles.size());
                target = new Target(adjacentTiles.get(rand));
            }
            return target;
        }

    }

}
